/**
 * \file mcql/core/msrunpeptidelist.cpp
 * \date 25/09/2024
 * \author Olivier Langella
 * \brief msrun model including fragmented peptides in masschroqlite
 */

/*******************************************************************************
 * Copyright (c) 2024 Olivier Langella
 *<Olivier.Langella@universite-paris-saclay.fr>.
 *
 * This file is part of MassChroQ.
 *
 *     MassChroQ is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MassChroQ is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MassChroQ.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

#include "msrunpeptidelist.h"
#include <pappsomspp/xicextractor/msrunxicextractorfactory.h>
#include <QtConcurrent>

mcql::MsRunPeptideList::MsRunPeptideList(mcql::MsRunSp msrun)
{
  msp_msRun = msrun;
}

mcql::MsRunPeptideList::MsRunPeptideList(const mcql::MsRunPeptideList &other)
{
  msp_msRun = other.msp_msRun;
}

mcql::MsRunPeptideList::~MsRunPeptideList()
{
}

void
mcql::MsRunPeptideList::addPeptideSpectrumIndexObservation(
  mcql::PeptideSp peptide_sp, std::size_t spectrum_index, std::uint8_t charge)
{

  auto it = m_peptideObservationList.insert(
    {peptide_sp.get(), std::make_shared<mcql::PeptideObservation>(peptide_sp)});
  it.first->second.get()->addObservation(
    {spectrum_index,
     charge,
     msp_msRun.get()->getPrecursorSPtrBySpectrumIndex(spectrum_index),
     true});
}

void
mcql::MsRunPeptideList::addPeptideScanNumberObservation(
  mcql::PeptideSp peptide_sp, std::size_t scan_number, std::uint8_t charge)
{

  auto it = m_peptideObservationList.insert(
    {peptide_sp.get(), std::make_shared<mcql::PeptideObservation>(peptide_sp)});
  it.first->second.get()->addObservation(
    {scan_number,
     charge,
     msp_msRun.get()->getPrecursorSPtrByScanNumber(scan_number),
     false});
}

void
mcql::MsRunPeptideList::quantify(
  const QString &tmp_dir,
  double retention_time_range_seconds,
  pappso::UiMonitorInterface &ui_monitor,
  const QuantificationMethod &quantification_method,
  double ni_minimum_abundance)
{
  m_peptideMeasurementsList.clear();
  ui_monitor.setStatus(QString("Starting quantification for msrun %1")
                         .arg(msp_msRun.get()
                                ->getMsRunReaderSPtr()
                                ->getMsRunId()
                                .get()
                                ->getFileName()));

  qDebug();
  pappso::MsRunXicExtractorFactory::getInstance().setTmpDir(tmp_dir);

  qDebug();
  pappso::MsRunXicExtractorFactory::getInstance()
    .setMsRunXicExtractorFactoryType(
      pappso::MsRunXicExtractorFactoryType::diskbuffer);

  qDebug();
  ui_monitor.setStatus(QString("Preparing extraction list"));
  std::vector<pappso::XicCoordSPtr> xic_coord_list =
    buildXicCoordList(quantification_method, ni_minimum_abundance);

  ui_monitor.setStatus(QString("Preparing extraction list done"));

  ui_monitor.setStatus(QString("Building XIC extractor"));
  pappso::MsRunXicExtractorInterfaceSp extractor_pwiz =
    pappso::MsRunXicExtractorFactory::getInstance().buildMsRunXicExtractorSp(
      msp_msRun.get()->getMsRunReaderSPtr());

  qDebug();
  extractor_pwiz->setXicExtractMethod(
    quantification_method.getXicExtractMethod());

  qDebug();

  extractor_pwiz.get()->setRetentionTimeAroundTarget(
    retention_time_range_seconds);

  qDebug();
  pappso::FilterInterfaceCstSPtr csp_filter_suite =
    std::make_shared<const pappso::FilterSuite>(
      quantification_method.getXicFilters());

  qDebug();
  extractor_pwiz.get()->setPostExtractionTraceFilterCstSPtr(csp_filter_suite);

  qDebug();

  ui_monitor.setStatus(QString("Building XIC extractor done"));
  ui_monitor.setStatus(QString("Extracting"));
  extractor_pwiz.get()->extractXicCoordSPtrListParallelized(ui_monitor,
                                                            xic_coord_list);

  qDebug();

  detectQuantifyPeaks(ui_monitor, quantification_method);


  ui_monitor.setStatus(QString("Quantification for msrun %1 done")
                         .arg(msp_msRun.get()
                                ->getMsRunReaderSPtr()
                                ->getMsRunId()
                                .get()
                                ->getFileName()));
}


std::vector<pappso::XicCoordSPtr>
mcql::MsRunPeptideList::buildXicCoordList(
  const QuantificationMethod &quantification_method,
  double ni_minimum_abundance)
{
  std::vector<pappso::XicCoordSPtr> xic_coord_list;

  for(auto pair_peptide : m_peptideObservationList)
    {
      mcql::PeptideObservationSp peptide_events_sp = pair_peptide.second;

      addPeptideObservation2XicCoordList(quantification_method,
                                         pair_peptide.second,
                                         xic_coord_list,
                                         ni_minimum_abundance);
      // peptide_events_sp.get()
    }


  return xic_coord_list;
}

void
mcql::MsRunPeptideList::addPeptideObservation2XicCoordList(
  const QuantificationMethod &quantification_method,
  const mcql::PeptideObservationSp &peptide_events_sp,
  std::vector<pappso::XicCoordSPtr> &xic_coord_list,
  double ni_minimum_abundance)
{
  PeptideMeasurementsSp peptide_measurements =
    std::make_shared<mcql::PeptideMeasurements>(peptide_events_sp);

  m_peptideMeasurementsList.push_back(peptide_measurements);

  peptide_measurements.get()->prepareMeasurements(quantification_method,
                                                  ni_minimum_abundance);
  peptide_measurements.get()->pushBackXicCoordList(xic_coord_list);
}

void
mcql::MsRunPeptideList::detectQuantifyPeaks(
  pappso::UiMonitorInterface &m_uiMonitor,
  const QuantificationMethod &quantification_method)
{

  m_uiMonitor.setStatus(QString("Detect and quantify peak"));
  /*
    for(auto &measure : m_peptideMeasurementsList)
      {
        measure->detectQuantifyPeaks(quantification_method);
      }
  */

  std::function<void(const mcql::PeptideMeasurementsSp &)>
    mapdetectQuantifyPeaks =
      [quantification_method](const mcql::PeptideMeasurementsSp &measure) {
        measure->detectQuantifyPeaks(quantification_method);
      };


  QFuture<void> res =
    QtConcurrent::map<std::vector<mcql::PeptideMeasurementsSp>::iterator>(
      m_peptideMeasurementsList.begin(),
      m_peptideMeasurementsList.end(),
      mapdetectQuantifyPeaks);
  res.waitForFinished();


  m_uiMonitor.setStatus(QString("Detect and quantify peak done"));
}

void
mcql::MsRunPeptideList::writeQuantificationResults(
  pappso::UiMonitorInterface &ui_monitor,
  mcql::QuantificationResultWriterInterfaceSp writer_sp,
  const QString &quantify_id,
  const QString &withingroup) const
{
  writer_sp.get()->setMsRunSp(msp_msRun);
  writer_sp.get()->setQuantificationId(quantify_id);
  writer_sp.get()->setGroupId(withingroup);

  ui_monitor.setStatus(
    QString("Writing quantification results for %1 %2 msrun %3")
      .arg(quantify_id)
      .arg(withingroup)
      .arg(msp_msRun.get()
             ->getMsRunReaderSPtr()
             .get()
             ->getMsRunId()
             .get()
             ->getSampleName()));

  ui_monitor.setTotalSteps(m_peptideMeasurementsList.size());
  writer_sp.get()->startPeptideMeasurements();
  for(const mcql::PeptideMeasurementsSp &measures : m_peptideMeasurementsList)
    {
      writer_sp.get()->write(*(measures.get()));
      ui_monitor.count();
    }
  ui_monitor.setTotalSteps(0);
  writer_sp.get()->endPeptideMeasurements();
}

const mcql::MsRunSp &
mcql::MsRunPeptideList::getMsRunSp() const
{
  return msp_msRun;
}

void
mcql::MsRunPeptideList::flushChromatogramTraces()
{

  for(auto &measure : m_peptideMeasurementsList)
    {
      measure.get()->flushXics();
    }
}
