/**
 * \file mcql/core/msrunpeptidelist.h
 * \date 25/09/2024
 * \author Olivier Langella
 * \brief msrun model including fragmented peptides in masschroqlite
 */

/*******************************************************************************
 * Copyright (c) 2024 Olivier Langella
 *<Olivier.Langella@universite-paris-saclay.fr>.
 *
 * This file is part of MassChroQ.
 *
 *     MassChroQ is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MassChroQ is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MassChroQ.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

#pragma once


#include "msrun.h"
#include "peptideobservation.h"
#include "peptidemeasurements.h"
#include <cstdint>
#include "../lib/quantifications/quantificationMethod.h"
#include "../output/quantificationresultwriterinterface.h"

namespace mcql
{

class MsRunPeptideList;

typedef std::shared_ptr<MsRunPeptideList> MsRunPeptideListSp;

/**
 * group together an msrun, available peptide observations, and peptide
 * measurements
 */
class MsRunPeptideList
{
  public:
  /**
   * Default constructor
   */
  MsRunPeptideList(MsRunSp msrun);

  /**
   * Copy constructor
   *
   * @param other TODO
   */
  MsRunPeptideList(const MsRunPeptideList &other);

  /**
   * Destructor
   */
  virtual ~MsRunPeptideList();

  void addPeptideSpectrumIndexObservation(mcql::PeptideSp peptide_sp,
                                          std::size_t spectrum_index,
                                          std::uint8_t charge);
  void addPeptideScanNumberObservation(mcql::PeptideSp peptide_sp,
                                       std::size_t spectrum_index,
                                       std::uint8_t charge);

  void quantify(const QString &tmp_dir,
                double retention_time_range_seconds,
                pappso::UiMonitorInterface &m_uiMonitor,
                const QuantificationMethod &quantification_method,
                double ni_minimum_abundance);

  void writeQuantificationResults(
    pappso::UiMonitorInterface &ui_monitor,
    mcql::QuantificationResultWriterInterfaceSp writer_sp,
    const QString &quantify_id,
    const QString &withingroup) const;

  const MsRunSp &getMsRunSp() const;

  /** flush (remove) transient chromatogram traces after peak detection to save space in
   * memory
   */
  void flushChromatogramTraces();

  private:
  std::vector<pappso::XicCoordSPtr>
  buildXicCoordList(const QuantificationMethod &quantification_method,
                    double ni_minimum_abundance);

  void detectQuantifyPeaks(pappso::UiMonitorInterface &m_uiMonitor,
                           const QuantificationMethod &quantification_method);

  void addPeptideObservation2XicCoordList(
    const QuantificationMethod &quantification_method,
    const mcql::PeptideObservationSp &peptide_events_sp,
    std::vector<pappso::XicCoordSPtr> &xic_coord_list,
    double ni_minimum_abundance);

  private:
  MsRunSp msp_msRun;

  std::map<mcql::Peptide *, mcql::PeptideObservationSp>
    m_peptideObservationList;

  std::vector<mcql::PeptideMeasurementsSp> m_peptideMeasurementsList;
};
} // namespace mcql
