/**
 * \file mcql/core/peptideobservation.cpp
 * \date 24/09/2024
 * \author Olivier Langella
 * \brief peptide model in masschroqlite
 */

/*******************************************************************************
 * Copyright (c) 2024 Olivier Langella
 *<Olivier.Langella@universite-paris-saclay.fr>.
 *
 * This file is part of MassChroQ.
 *
 *     MassChroQ is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MassChroQ is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MassChroQ.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/


#include "peptideobservation.h"
#include <pappsomspp/pappsoexception.h>
namespace mcql
{

PeptideObservation::PeptideObservation(mcql::PeptideSp peptide)
  : msp_peptide(peptide)
{
}


PeptideObservation::PeptideObservation(const PeptideObservation &other)
  : msp_peptide(other.msp_peptide)
{
  m_observationList = other.m_observationList;
}

PeptideObservation::~PeptideObservation()
{
}

void
PeptideObservation::addObservation(
  const mcql::PeptideObservation::Observation &observation)
{
  m_observationList.push_back(observation);
}
} // namespace mcql

std::vector<std::uint8_t>
mcql::PeptideObservation::getObservedChargeStates() const
{
  std::vector<std::uint8_t> charge_list;
  for(auto &observation : m_observationList)
    {
      charge_list.push_back(observation.m_charge);
    }
  std::sort(charge_list.begin(), charge_list.end());

  charge_list.erase(std::unique(charge_list.begin(), charge_list.end()),
                    charge_list.end());
  return charge_list;
}

const mcql::PeptideSp &
mcql::PeptideObservation::getPeptideSp() const
{
  return msp_peptide;
}

double
mcql::PeptideObservation::getBestRtOverallChargeStates() const
{
  double best_rt   = 0;
  double intensity = -1;
  for(auto &observation : m_observationList)
    {
      if(observation.msp_precursor.get()->getIntensity() > intensity)
        {
          best_rt =
            observation.msp_precursor.get()->getXicCoordSPtr().get()->rtTarget;
          intensity = observation.msp_precursor.get()->getIntensity();
        }
    }

  return best_rt;
}

const pappso::XicCoordSPtr &
mcql::PeptideObservation::getXicCoordSPtrForCharge(std::uint8_t charge) const
{

  // get XIC coordinates of the most intense precursor
  std::size_t observation_indice  = 0;
  pappso::pappso_double intensity = -1;

  std::size_t i = 0;
  for(auto &observation : m_observationList)
    {
      if(observation.m_charge == charge)
        {
          if(observation.msp_precursor.get()->getIntensity() > intensity)
            {
              observation_indice = i;
              intensity = observation.msp_precursor.get()->getIntensity();
              qDebug();
            }
        }
      i++;
    }

  return m_observationList.at(observation_indice)
    .msp_precursor.get()
    ->getXicCoordSPtr();
}

std::vector<double>
mcql::PeptideObservation::getObservedRetentionTimesOverallChargeStates() const
{
  std::vector<double> rt_list;
  for(auto &observation : m_observationList)
    {
      rt_list.push_back(
        observation.msp_precursor.get()->getXicCoordSPtr().get()->rtTarget);
    }

  return rt_list;
}
