/**
 * \file mcql/input/masschroqmlparser.cpp
 * \date 23/09/2024
 * \author Olivier Langella
 * \brief masschroqML parser for MassChroqLight
 */

/*******************************************************************************
 * Copyright (c) 2024 Olivier Langella
 *<Olivier.Langella@universite-paris-saclay.fr>.
 *
 * This file is part of MassChroQ.
 *
 *     MassChroQ is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MassChroQ is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MassChroQ.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

#include "masschroqmlparser.h"
#include <QDir>
#include <QFileInfo>
#include <pappsomspp/pappsoexception.h>
#include <pappsomspp/exception/exceptionnotrecognized.h>
#include <pappsomspp/msfile/msfileaccessor.h>
#include <pappsomspp/processing/filters/filterinterface.h>
#include <pappsomspp/processing/filters/filtermorpho.h>
#include <pappsomspp/processing/filters/filterpass.h>
#include <pappsomspp/processing/detection/tracedetectionzivy.h>
#include <pappsomspp/processing/detection/tracedetectionmoulon.h>
#include "../output/quantificationresultodsoutput.h"
#include <QtConcurrent>

namespace mcql
{

MassChroqMlParser::MassChroqMlParser(pappso::UiMonitorInterface &ui_monitor,
                                     const QString &tmp_dir_name)
  : m_tmpDirName(tmp_dir_name), m_uiMonitor(ui_monitor)
{
}

MassChroqMlParser::~MassChroqMlParser()
{
}
bool
MassChroqMlParser::readFile(const QString &fileName)
{

  QDir::setCurrent(QFileInfo(fileName).absolutePath());
  return pappso::XmlStreamReaderInterface::readFile(fileName);
}

void
mcql::MassChroqMlParser::readStream()
{

  try
    {
      if(m_qxmlStreamReader.readNextStartElement())
        {
          if(m_qxmlStreamReader.name().toString() == "masschroq")
            {

              while(m_qxmlStreamReader.readNextStartElement())
                {
                  qDebug() << m_qxmlStreamReader.name();
                  if(m_qxmlStreamReader.name().toString() ==
                     "project_parameters")
                    readProjectParameters();
                  else if(m_qxmlStreamReader.name().toString() == "rawdata")
                    readRawdata();
                  else if(m_qxmlStreamReader.name().toString() == "groups")
                    readGroups();
                  else if(m_qxmlStreamReader.name().toString() ==
                          "protein_list")
                    readProteinList();
                  else if(m_qxmlStreamReader.name().toString() ==
                          "peptide_list")
                    readPeptideList();
                  else if(m_qxmlStreamReader.name().toString() == "alignments")
                    readAlignments();
                  else if(m_qxmlStreamReader.name().toString() ==
                          "quantification_methods")
                    readQuantificationMethods();
                  else if(m_qxmlStreamReader.name().toString() ==
                          "quantification")
                    readQuantification();
                  else
                    {
                      m_qxmlStreamReader.raiseError(
                        QObject::tr(
                          "ERROR parsing XML file at tag %1:\n%2 unknown")
                          .arg(m_qxmlStreamReader.name()));
                    }
                }
            }
        }
    }

  catch(const pappso::PappsoException &e)
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("ERROR parsing XML file at tag %1:\n%2")
          .arg(m_qxmlStreamReader.name())
          .arg(e.qwhat()));
    }
}

void
MassChroqMlParser::readProjectParameters()
{
  //<project_param category="1" name="i2MassChroQ_VERSION" value="1.0.9"/>
  m_qxmlStreamReader.skipCurrentElement();
}


void
MassChroqMlParser::readAlignments()
{
  m_qxmlStreamReader.skipCurrentElement();
}

void
MassChroqMlParser::readQuantificationMethods()
{

  while(m_qxmlStreamReader.readNextStartElement())
    {
      qDebug() << m_qxmlStreamReader.name();
      read_quantification_method();
    }
}

void
MassChroqMlParser::readQuantification()
{

  while(m_qxmlStreamReader.readNextStartElement())
    {
      qDebug() << m_qxmlStreamReader.name();

      if(m_qxmlStreamReader.name().toString() == "quantification_results")
        read_quantification_results();
      else if(m_qxmlStreamReader.name().toString() == "quantification_traces")
        m_qxmlStreamReader.skipCurrentElement();
      else if(m_qxmlStreamReader.name().toString() == "quantify")
        read_quantify();

      else
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("ERROR parsing XML file at tag %1:\n%2 unknown")
              .arg(m_qxmlStreamReader.name()));
        }
    }
}

void
MassChroqMlParser::readRawdata()
{
  std::map<QString, QString> msfilepathlist;
  while(m_qxmlStreamReader.readNextStartElement())
    {
      qDebug() << m_qxmlStreamReader.name();
      //<data_file id="samp0" format="mzxml" path="bsa1.mzXML"
      // type="profile" />
      msfilepathlist.insert(read_data_file());
    }


  std::function<mcql::MsRunPeptideListSp(const std::pair<QString, QString> &)>
    mapFilenameList = [](const std::pair<QString, QString> &mapit) {
      pappso::MsFileAccessor file_access(mapit.second, mapit.first);
      file_access.setPreferredFileReaderType(pappso::MsDataFormat::brukerTims,
                                             pappso::FileReaderType::tims_ms2);

      std::vector<pappso::MsRunIdCstSPtr> msrunid_list =
        file_access.getMsRunIds();


      if(file_access.getFileFormat() == pappso::MsDataFormat::unknown)
        {
          throw pappso::ExceptionNotRecognized(
            "MS data file format not recognized");
        }
      if(msrunid_list.size() == 0)
        {
          throw pappso::PappsoException("msrunid_list.size() == 0");
        }


      pappso::MsRunReaderSPtr run_reader =
        file_access.getMsRunReaderSPtrByRunId("", mapit.first);
      run_reader.get()->setMonoThread(true);
      mcql::MsRunSp msrun = std::make_shared<mcql::MsRun>(run_reader);
      run_reader.get()->releaseDevice();
      run_reader.get()->setMonoThread(false);
      return std::make_shared<mcql::MsRunPeptideList>(msrun);
    };


  m_uiMonitor.setTitle(
    QObject::tr("reading %1 msruns").arg(msfilepathlist.size()));


  pappso::UiMonitorInterface *local_monitor = &m_uiMonitor;
  std::map<QString, mcql::MsRunPeptideListSp> *p_localPtrOnmsrunpeptidelist =
    &m_msfileList;

  std::function<void(std::size_t, const mcql::MsRunPeptideListSp)>
    reduce_function =
      [local_monitor, p_localPtrOnmsrunpeptidelist](
        std::size_t result, const mcql::MsRunPeptideListSp msrun) {
        local_monitor->setStatus(QObject::tr("MS run '%1' from file %2: added")
                                   .arg(msrun.get()
                                          ->getMsRunSp()
                                          .get()
                                          ->getMsRunReaderSPtr()
                                          .get()
                                          ->getMsRunId()
                                          .get()
                                          ->getXmlId())
                                   .arg(msrun.get()
                                          ->getMsRunSp()
                                          .get()
                                          ->getMsRunReaderSPtr()
                                          .get()
                                          ->getMsRunId()
                                          .get()
                                          ->getFileName()));

        p_localPtrOnmsrunpeptidelist->insert({msrun.get()
                                                ->getMsRunSp()
                                                .get()
                                                ->getMsRunReaderSPtr()
                                                .get()
                                                ->getMsRunId()
                                                .get()
                                                ->getXmlId(),
                                              msrun});
        result++;
      };

  QFuture<std::size_t> res =
    QtConcurrent::mappedReduced<std::size_t>(msfilepathlist.begin(),
                                             msfilepathlist.end(),
                                             mapFilenameList,
                                             reduce_function,
                                             QtConcurrent::OrderedReduce);
  res.waitForFinished();
}


std::pair<QString, QString>
MassChroqMlParser::read_data_file()
{
  std::pair<QString, QString> id_file;
  if(m_qxmlStreamReader.name().toString() == "data_file")
    {
      QString filename =
        m_qxmlStreamReader.attributes().value("path").toString();
      if(filename.isEmpty())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("The data_file tag must have a path attribute."));
        }

      QString idname = m_qxmlStreamReader.attributes().value("id").toString();
      if(idname.isEmpty())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("The data_file tag must have an id attribute."));
        }

      QFileInfo filenameInfo(filename);
      if(!filenameInfo.exists())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("cannot find input file : %2 \n").arg(filename));
        }
      else if(!filenameInfo.isReadable())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("cannot read input file : %2 \n").arg(filename));
        }

      QString type = m_qxmlStreamReader.attributes().value("type").toString();
      if(!type.isEmpty() && type == "srm")
        {
          m_qxmlStreamReader.raiseError(QObject::tr(
            "The SRM mode is not yet implemented in MassChroQ, but it will: "
            "\n%1"));
        }

      try
        {

          id_file = std::pair<QString, QString>(
            {idname, filenameInfo.absoluteFilePath()});

          m_qxmlStreamReader.skipCurrentElement();
        }
      catch(pappso::PappsoException &pappso_error)
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("problem creating msrun %1:\n%2")
              .arg(idname, pappso_error.qwhat()));
        }
    }

  else
    {
      m_qxmlStreamReader.raiseError(QObject::tr("Not a data_file element"));
      m_qxmlStreamReader.skipCurrentElement();
    }
  return id_file;
}


void
MassChroqMlParser::readGroups()
{
  qDebug();

  if(m_qxmlStreamReader.name().toString() == "groups")
    {

      while(m_qxmlStreamReader.readNextStartElement())
        {
          qDebug() << m_qxmlStreamReader.name();
          //<group id="all_msrun" data_ids="msruna1 msruna2 msruna3 msruna4
          // msruna5 msruna6 msruna7 msruna8 msruna9 msrunb10 msrunb11 msrunb12
          // msrunb13 msrunb14 msrunb15 msrunb16 msrunb17 msrunb18 msrunb19"/>

          read_group();
        }
    }
  else
    {
      m_qxmlStreamReader.raiseError(QObject::tr("Not a groups element"));
      m_qxmlStreamReader.skipCurrentElement();
    }
}
void
MassChroqMlParser::readProteinList()
{
  qDebug();
  qDebug() << m_qxmlStreamReader.name();
  if(m_qxmlStreamReader.name().toString() == "protein_list")
    {

      while(m_qxmlStreamReader.readNextStartElement())
        {
          qDebug() << m_qxmlStreamReader.name();
          //<protein desc="conta|P02769|ALBU_BOVIN SERUM ALBUMIN PRECURSOR."
          // id="P1.1" />
          read_protein();
        }
    }
  else
    {
      m_qxmlStreamReader.raiseError(QObject::tr("Not a protein_list element"));
      m_qxmlStreamReader.skipCurrentElement();
    }
}

void
MassChroqMlParser::readPeptideList()
{
  qDebug();


  m_uiMonitor.setTitle("Reading peptide list");
  while(m_qxmlStreamReader.readNextStartElement())
    {
      qDebug() << m_qxmlStreamReader.name();
      read_peptide();
    }
}


void
MassChroqMlParser::read_peptide()
{

  mcql::PeptideSp peptide_sp;
  qDebug() << m_qxmlStreamReader.name();
  QStringList msrun_id_list;
  if(m_qxmlStreamReader.name().toString() == "peptide")
    {
      QString idname = m_qxmlStreamReader.attributes().value("id").toString();
      QString mh     = m_qxmlStreamReader.attributes().value("mh").toString();
      QString seq    = m_qxmlStreamReader.attributes().value("seq").toString();
      QString mods   = m_qxmlStreamReader.attributes().value("mods").toString();

      std::vector<mcql::ProteinSp> protein_list;

      // prot_ids
      QStringList st_prot_list =
        m_qxmlStreamReader.attributes()
          .value("prot_ids")
          .toString()
          .split(" ", Qt::SplitBehaviorFlags::SkipEmptyParts);
      if(st_prot_list.size() == 0)
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("the peptide tag must have a prot_ids attribute."));
        }
      else
        {
          for(const QString &prot_id : st_prot_list)
            {
              auto it = m_proteinMap.find(prot_id);
              if(it != m_proteinMap.end())
                {
                  protein_list.push_back(it->second);
                }
              else
                {
                  m_qxmlStreamReader.raiseError(
                    QObject::tr("protein id %1 not found").arg(prot_id));
                }
            }
        }
      if(idname.isEmpty())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("the peptide tag must have an id attribute."));
        }
      if(mh.isEmpty())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("the peptide tag must have a mh attribute."));
        }
      /// create a new Peptide object and set its members
      if(seq.isEmpty())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("the peptide tag must have a seq attribute."));
        }
      else
        {
          try
            {
              pappso::Peptide peptide(seq);

              // modifications + observed_in
              while(m_qxmlStreamReader.readNextStartElement())
                {
                  qDebug() << m_qxmlStreamReader.name();
                  if(m_qxmlStreamReader.name().toString() == "modifications")
                    {
                      while(m_qxmlStreamReader.readNextStartElement())
                        {
                          if(m_qxmlStreamReader.name().toString() == "psimod")
                            {
                              read_psimod(peptide);
                            }
                          else
                            {
                              m_qxmlStreamReader.raiseError(
                                QObject::tr("%1 element not allowed here")
                                  .arg(m_qxmlStreamReader.name()));
                              m_qxmlStreamReader.skipCurrentElement();
                            }
                        }
                    }
                  else if(m_qxmlStreamReader.name().toString() == "observed_in")
                    {
                      if(peptide_sp.get() == nullptr)
                        {
                          peptide_sp = std::make_shared<mcql::Peptide>(
                            idname, peptide.makePeptideSp(), protein_list);
                          m_peptideStore.push_back(peptide_sp);
                          peptide_sp.get()->setMods(mods);
                        }

                      read_observed_in(peptide_sp, msrun_id_list);
                    }
                  else
                    {
                      m_qxmlStreamReader.raiseError(
                        QObject::tr("%1 element not allowed here")
                          .arg(m_qxmlStreamReader.name()));
                      m_qxmlStreamReader.skipCurrentElement();
                    }
                }
            }
          catch(const pappso::PappsoException &e)
            {
              m_qxmlStreamReader.raiseError(
                QObject::tr("error building peptide %1\n%2")
                  .arg(seq)
                  .arg(e.qwhat()));
            }
        }
    }
  else
    {
      m_qxmlStreamReader.raiseError(QObject::tr("Not a peptide element"));
      m_qxmlStreamReader.skipCurrentElement();
    }

  if(peptide_sp.get() == nullptr)
    {
      // mark for absence of this peptide in msruns :
      for(auto &key_file : m_msfileList)
        {
          QString msrun_id = key_file.first;
          if(!msrun_id_list.contains(msrun_id))
            {
              // no observation in this msrun
              peptide_sp.get()->addNotObservedInMsRunSp(
                key_file.second.get()->getMsRunSp());
            }
        }
    }
  qDebug() << m_qxmlStreamReader.name();
}


void
MassChroqMlParser::read_psimod(pappso::Peptide &peptide)
{

  qDebug() << m_qxmlStreamReader.name();
  QString at  = m_qxmlStreamReader.attributes().value("at").toString();
  QString acc = m_qxmlStreamReader.attributes().value("acc").toString();
  if(at.isEmpty())
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("the psimod tag must have an at attribute."));
    }
  if(acc.isEmpty())
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("the psimod tag must have an acc attribute."));
    }
  try
    {
      pappso::AaModificationP modification =
        pappso::AaModification::getInstance(acc);
      peptide.addAaModification(modification, at.toUInt() - 1);
    }
  catch(const pappso::PappsoException &e)
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("ERROR setting psimod:\n%1").arg(e.qwhat()));
    }

  m_qxmlStreamReader.skipCurrentElement();
}


/// <observed_in data="delumeau1" scan="33" z="2">
void
MassChroqMlParser::read_observed_in(const mcql::PeptideSp &peptide_sp,
                                    QStringList &msrun_id_list)
{
  qDebug() << m_qxmlStreamReader.name();
  if(m_qxmlStreamReader.name().toString() == "observed_in")
    {
      QString data  = m_qxmlStreamReader.attributes().value("data").toString();
      QString scan  = m_qxmlStreamReader.attributes().value("scan").toString();
      QString index = m_qxmlStreamReader.attributes().value("index").toString();
      QString z     = m_qxmlStreamReader.attributes().value("z").toString();
      if(data.isEmpty())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("the observed_in tag must have a data attribute "
                        "(reference on a valid msRun id)."));
        }
      if(scan.isEmpty() && index.isEmpty())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("the observed_in tag must have an index attribute."));
        }
      if(z.isEmpty())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("the observed_in tag must have a z attribute."));
        }

      auto it = m_msfileList.find(data);
      if(it == m_msfileList.end())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("the observed_in tag contains a data attribute "
                        "that does not reference any msrun..."));
        }
      if(!msrun_id_list.contains(data))
        {
          msrun_id_list << data;
        }
      /// add the observed in infos to the current peptide
      mcql::MsRunPeptideListSp ms_run = it->second;
      try
        {
          if(scan.isEmpty())
            {

              ms_run.get()->addPeptideSpectrumIndexObservation(
                peptide_sp,
                (std::size_t)index.toULongLong(),
                (std::uint8_t)z.toUInt());
            }
          else
            {

              ms_run.get()->addPeptideScanNumberObservation(
                peptide_sp,
                (std::size_t)scan.toULongLong(),
                (std::uint8_t)z.toUInt());
            }

          peptide_sp.get()->addObservedChargeState((std::uint8_t)z.toUInt());
        }
      catch(pappso::PappsoException &err)
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("error reading observed_in tag : %1").arg(err.qwhat()));
        }

      m_qxmlStreamReader.skipCurrentElement();
    }

  else
    {
      m_qxmlStreamReader.raiseError(QObject::tr("Not an observed_in element"));
      m_qxmlStreamReader.skipCurrentElement();
    }
}


/// <quantification_method id="my_qzivy">
void
MassChroqMlParser::read_quantification_method()
{

  qDebug() << m_qxmlStreamReader.name();
  if(m_qxmlStreamReader.name().toString() == "quantification_method")
    {
      QString quantification_id =
        m_qxmlStreamReader.attributes().value("id").toString();
      if(quantification_id.isEmpty())
        {
          m_qxmlStreamReader.raiseError(QObject::tr(
            "the quantification_method tag must have an id attribute."));
        }

      QuantificationMethod quantification_method(quantification_id);

      // xic_extraction
      m_qxmlStreamReader.readNextStartElement();
      if(m_qxmlStreamReader.name().toString() != "xic_extraction")
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("xic_extraction element expected here"));
        }
      else
        {
          QString xic_type =
            m_qxmlStreamReader.attributes().value("xic_type").toString();
          if(xic_type == "sum")
            {
              quantification_method.setXicExtractMethod(
                pappso::XicExtractMethod::sum);
            }
          else if(xic_type == "max")
            {
              quantification_method.setXicExtractMethod(
                pappso::XicExtractMethod::max);
            }
          else
            {
              m_qxmlStreamReader.raiseError(
                QObject::tr("the xic_extraction tag must have a xic_type "
                            "attribute whose value is 'sum' or 'max'."));
            }

          if(m_qxmlStreamReader.readNextStartElement())
            {
              // mz_range
              if(m_qxmlStreamReader.name().toString() == "mz_range")
                {
                  /// <xic_extraction>
                  /// <mz_range min="0.5" max="1.5"/>
                  QString min =
                    m_qxmlStreamReader.attributes().value("min").toString();
                  QString max =
                    m_qxmlStreamReader.attributes().value("max").toString();
                  if(min.isEmpty())
                    {
                      m_qxmlStreamReader.raiseError(QObject::tr(
                        "the mz_range tag must have a min attribute."));
                    }
                  if(max.isEmpty())
                    {
                      m_qxmlStreamReader.raiseError(QObject::tr(
                        "the mz_range tag must have a max attribute."));
                    }

                  quantification_method.setXicExtractPrecisionUnit(
                    pappso::PrecisionUnit::mz);
                  quantification_method.setXicExtractLowerPrecision(
                    min.toDouble());
                  quantification_method.setXicExtractUpperPrecision(
                    max.toDouble());
                  m_qxmlStreamReader.skipCurrentElement();
                }
              else if(m_qxmlStreamReader.name().toString() == "ppm_range")
                {
                  /// <xic_extraction>
                  /// <ppm_range min="" max=""/>

                  QString min =
                    m_qxmlStreamReader.attributes().value("min").toString();
                  QString max =
                    m_qxmlStreamReader.attributes().value("max").toString();
                  if(min.isEmpty())
                    {
                      m_qxmlStreamReader.raiseError(QObject::tr(
                        "the ppm_range tag must have a min attribute."));
                    }
                  if(max.isEmpty())
                    {
                      m_qxmlStreamReader.raiseError(QObject::tr(
                        "the ppm_range tag must have a max attribute."));
                    }

                  quantification_method.setXicExtractPrecisionUnit(
                    pappso::PrecisionUnit::ppm);
                  quantification_method.setXicExtractLowerPrecision(
                    min.toDouble());
                  quantification_method.setXicExtractUpperPrecision(
                    max.toDouble());
                  m_qxmlStreamReader.skipCurrentElement();
                }
              else
                {
                  m_qxmlStreamReader.raiseError(
                    QObject::tr("missing mz_range or ppm_range element"));
                }
            }

          else
            {
              m_qxmlStreamReader.raiseError(
                QObject::tr("missing mz_range or ppm_range element"));
            }

          if(m_qxmlStreamReader.readNextStartElement())
            {
              m_qxmlStreamReader.raiseError(
                QObject::tr("%1 element not allowed here")
                  .arg(m_qxmlStreamReader.name()));
            }

          qDebug() << m_qxmlStreamReader.name();
        }
      qDebug() << m_qxmlStreamReader.name();


      m_qxmlStreamReader.readNextStartElement();
      // xic_filters
      if(m_qxmlStreamReader.name().toString() == "xic_filters")
        {
          read_xic_filters(quantification_method);
          // peak_detection
          m_qxmlStreamReader.readNextStartElement();
        }

      qDebug() << m_qxmlStreamReader.name();
      read_peak_detection(quantification_method);

      qDebug() << m_qxmlStreamReader.name();
      m_quantificationMethodList.insert(
        {quantification_id, quantification_method});
      //_p_current_quantification_method = NULL;
      qDebug() << m_qxmlStreamReader.name();
      if(m_qxmlStreamReader.readNextStartElement())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("%1 element not allowed here")
              .arg(m_qxmlStreamReader.name()));
        }
      qDebug() << m_qxmlStreamReader.name();
    }

  else
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("Not a quantification_method element"));
      m_qxmlStreamReader.skipCurrentElement();
    }
}

void
MassChroqMlParser::read_peak_detection(
  QuantificationMethod &quantification_method)
{
  qDebug() << m_qxmlStreamReader.name();
  if(m_qxmlStreamReader.name().toString() == "peak_detection")
    {
      if(m_qxmlStreamReader.readNextStartElement())
        {

          if(m_qxmlStreamReader.name().toString() == "detection_zivy")
            {
              read_detection_zivy(quantification_method);
            }
          else if(m_qxmlStreamReader.name().toString() == "detection_moulon")
            {
              read_detection_moulon(quantification_method);
            }

          else
            {
              m_qxmlStreamReader.raiseError(QObject::tr(
                "detection_zivy or detection_moulon element expected"));
              m_qxmlStreamReader.skipCurrentElement();
            }
        }
      else
        {

          m_qxmlStreamReader.raiseError(
            QObject::tr("missing detection_zivy or detection_moulon element"));
        }


      if(m_qxmlStreamReader.readNextStartElement())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("%1 element not allowed here")
              .arg(m_qxmlStreamReader.name()));
        }
      qDebug() << m_qxmlStreamReader.name();
    }

  else
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("Not a peak_detection element"));
      m_qxmlStreamReader.skipCurrentElement();
    }
}

void
MassChroqMlParser::read_xic_filters(QuantificationMethod &quantification_method)
{
  qDebug() << m_qxmlStreamReader.name();
  if(m_qxmlStreamReader.name().toString() == "xic_filters")
    {
      while(m_qxmlStreamReader.readNextStartElement())
        {
          read_xic_filter(quantification_method);
        }
    }

  else
    {
      m_qxmlStreamReader.raiseError(QObject::tr("Not a xic_filters element"));
      m_qxmlStreamReader.skipCurrentElement();
    }
}

void
MassChroqMlParser::read_xic_filter(QuantificationMethod &quantification_method)
{
  qDebug() << m_qxmlStreamReader.name();
  if(m_qxmlStreamReader.name().toString() == "anti_spike")
    {
      /// <anti_spike half="8">
      QString half = m_qxmlStreamReader.attributes().value("half").toString();
      if(half.isEmpty())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("the anti_spike tag must have a half attribute."));
        }

      // pappso::FilterMorphoAntiSpike f_spike(half.toDouble());
      quantification_method.addFilter(
        std::make_shared<pappso::FilterMorphoAntiSpike>(half.toDouble()));
    }
  else if(m_qxmlStreamReader.name().toString() == "background")
    {

      //<background half_mediane="5" half_min_max="20">
      QString half_mediane =
        m_qxmlStreamReader.attributes().value("half_mediane").toString();
      QString half_min_max =
        m_qxmlStreamReader.attributes().value("half_min_max").toString();

      if(half_mediane.isEmpty())
        {
          m_qxmlStreamReader.raiseError(QObject::tr(
            "the background tag must have a half_mediane attribute."));
        }
      if(half_min_max.isEmpty())
        {
          m_qxmlStreamReader.raiseError(QObject::tr(
            "the background tag must have a half_min_max attribute."));
        }

      pappso::FilterMorphoBackground f_bkg(half_mediane.toDouble(),
                                           half_min_max.toDouble());
      quantification_method.addFilter(
        std::make_shared<pappso::FilterMorphoBackground>(f_bkg));
    }

  else if(m_qxmlStreamReader.name().toString() == "remove_intensity")
    {
      QString value = m_qxmlStreamReader.attributes().value("value").toString();

      if(value.isEmpty())
        {
          m_qxmlStreamReader.raiseError(QObject::tr(
            "the remove_intensity tag must have a value attribute."));
        }

      pappso::FilterRemoveY f_remove(value.toDouble());
      quantification_method.addFilter(
        std::make_shared<pappso::FilterRemoveY>(f_remove));
    }

  else if(m_qxmlStreamReader.name().toString() == "remove_quantile")
    {
      QString value = m_qxmlStreamReader.attributes().value("value").toString();

      if(value.isEmpty())
        {
          m_qxmlStreamReader.raiseError(QObject::tr(
            "the remove_quantile tag must have a value attribute."));
        }

      pappso::FilterQuantileBasedRemoveY f_remove(value.toDouble());
      quantification_method.addFilter(
        std::make_shared<pappso::FilterQuantileBasedRemoveY>(f_remove));
    }

  else if(m_qxmlStreamReader.name().toString() == "smoothing")
    {
      QString half = m_qxmlStreamReader.attributes().value("half").toString();
      if(half.isEmpty())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("the smoothing tag must have a half attribute."));
        }

      pappso::FilterMorphoMean f_smooth(half.toDouble());

      quantification_method.addFilter(
        std::make_shared<pappso::FilterMorphoMean>(f_smooth));
    }

  else
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("filter element, one of smoothing, remove_quantile, "
                    "remove_intensity, background, anti_spike expected"));
      m_qxmlStreamReader.skipCurrentElement();
    }
  m_qxmlStreamReader.skipCurrentElement();
}

/// <detection_zivy>
void
MassChroqMlParser::read_detection_zivy(
  QuantificationMethod &quantification_method)

{
  qDebug() << m_qxmlStreamReader.name();
  if(m_qxmlStreamReader.name().toString() == "detection_zivy")
    {
      std::shared_ptr<pappso::TraceDetectionZivy> sp_detection_zivy =
        std::make_shared<pappso::TraceDetectionZivy>(0, 2, 3, 1000, 100);

      while(m_qxmlStreamReader.readNextStartElement())
        {
          qDebug() << m_qxmlStreamReader.name();
          if(m_qxmlStreamReader.name().toString() == "mean_filter_half_edge")
            {
              /// <detection_zivy>
              /// <mean_filter_half_edge>3</mean_filter_half_edge>
              sp_detection_zivy.get()->setFilterMorphoMean(
                pappso::FilterMorphoMean(
                  m_qxmlStreamReader.readElementText().toDouble()));
            }
          else if(m_qxmlStreamReader.name().toString() == "minmax_half_edge")
            {
              /// <detection_zivy>
              /// <minmax_half_edge>3</minmax_half_edge>
              sp_detection_zivy.get()->setFilterMorphoMinMax(
                pappso::FilterMorphoMinMax(
                  m_qxmlStreamReader.readElementText().toDouble()));
            }
          else if(m_qxmlStreamReader.name().toString() == "maxmin_half_edge")
            {

              /// <detection_zivy>
              /// <maxmin_half_edge>3</maxmin_half_edge>
              sp_detection_zivy.get()->setFilterMorphoMaxMin(
                pappso::FilterMorphoMaxMin(
                  m_qxmlStreamReader.readElementText().toDouble()));
            }

          else if(m_qxmlStreamReader.name().toString() ==
                  "detection_threshold_on_max")
            {

              /// <detection_zivy>
              /// <detection_threshold_on_max>5000</detection_threshold_on_max>
              sp_detection_zivy.get()->setDetectionThresholdOnMinmax(
                m_qxmlStreamReader.readElementText().toDouble());
            }

          else if(m_qxmlStreamReader.name().toString() ==
                  "detection_threshold_on_min")
            {
              /// <detection_zivy>
              /// <detection_threshold_on_min>3000</detection_threshold_on_min>
              sp_detection_zivy.get()->setDetectionThresholdOnMaxmin(
                m_qxmlStreamReader.readElementText().toDouble());
            }

          else
            {
              m_qxmlStreamReader.raiseError(
                QObject::tr("detection_zivy element has no %1 element")
                  .arg(m_qxmlStreamReader.name()));
            }
        }


      qDebug() << m_qxmlStreamReader.name();
      quantification_method.setDetectionMethod(sp_detection_zivy);
    }

  else
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("Not a detection_zivy element"));
      m_qxmlStreamReader.skipCurrentElement();
    }
}

void
MassChroqMlParser::read_detection_moulon(
  QuantificationMethod &quantification_method)
{

  qDebug() << m_qxmlStreamReader.name();
  if(m_qxmlStreamReader.name().toString() == "detection_moulon")
    {
      /// <detection_moulon>
      std::shared_ptr<pappso::TraceDetectionMoulon>
        sp_current_detection_moulon =
          std::make_shared<pappso::TraceDetectionMoulon>(0, 100, 10);

      while(m_qxmlStreamReader.readNextStartElement())
        {
          qDebug() << m_qxmlStreamReader.name();
          if(m_qxmlStreamReader.name().toString() == "TIC_stop")
            {
              /// <detection_moulon>
              /// <TIC_stop>4</TIC_stop>
              sp_current_detection_moulon->setTicStop(
                m_qxmlStreamReader.readElementText().toDouble());
            }
          else if(m_qxmlStreamReader.name().toString() == "TIC_start")
            {
              /// <detection_moulon>
              /// <TIC_start>4</TIC_start>
              sp_current_detection_moulon->setTicStart(
                m_qxmlStreamReader.readElementText().toDouble());
            }
          else if(m_qxmlStreamReader.name().toString() == "smoothing_point")
            {
              /// <detection_moulon>
              /// <smoothing_point>5</smoothing_point>
              sp_current_detection_moulon->setFilterMorphoMean(
                m_qxmlStreamReader.readElementText().toDouble());
            }
          else
            {
              m_qxmlStreamReader.raiseError(
                QObject::tr("detection_moulon element has no %1 element")
                  .arg(m_qxmlStreamReader.name()));
            }
        }


      quantification_method.setDetectionMethod(
        std::make_shared<const pappso::TraceDetectionMoulon>(
          *sp_current_detection_moulon));
    }

  else
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("Not a detection_moulon element"));
      m_qxmlStreamReader.skipCurrentElement();
    }
}

void
MassChroqMlParser::read_quantify()
{
  qDebug() << m_qxmlStreamReader.name();
  if(m_qxmlStreamReader.name().toString() == "quantify")
    {
      QString quantify_id =
        m_qxmlStreamReader.attributes().value("id").toString();

      QString withingroup =
        m_qxmlStreamReader.attributes().value("withingroup").toString();
      /// <quantify id="q1" withingroup="G1" quantification_method_id="q1"
      /// compute_isotope_threshold="0.00001"> selection of masses to extract
      /// xic, detect peaks, quantify in this group
      QString quantification_method_id = m_qxmlStreamReader.attributes()
                                           .value("quantification_method_id")
                                           .toString();

      auto it = m_quantificationMethodList.find(quantification_method_id);
      if(it != m_quantificationMethodList.end())
        {
          readInQuantify(quantify_id, withingroup, it->second);
        }
    }
}

} // namespace mcql

void
mcql::MassChroqMlParser::readInQuantify(
  const QString &quantify_id,
  const QString &withingroup,
  QuantificationMethod &quantification_method)
{
  while(m_qxmlStreamReader.readNextStartElement())
    {
      qDebug() << m_qxmlStreamReader.name();
      if(m_qxmlStreamReader.name().toString() == "peptides_in_peptide_list")
        {
          //<peptides_in_peptide_list mode="post_matching"
          // isotope_label_refs="iso1 iso2" />
          //<peptides_in_peptide_list mode="post_matching"
          // ni_min_abundance="0.8"/>

          double ni_min_abundance = 0;
          if(m_qxmlStreamReader.attributes()
               .value("ni_min_abundance")
               .isEmpty())
            {
            }
          else
            {
              ni_min_abundance = m_qxmlStreamReader.attributes()
                                   .value("ni_min_abundance")
                                   .toDouble();
            }

          if(ni_min_abundance > 0)
            { /*
               for(auto &peptide_sp : m_peptideStore)
                 {
                   peptide_sp.get()->computeIsotopologues(ni_min_abundance);
                 }*/
              std::function<void(const mcql::PeptideSp &)>
                mapComputeIsotopologues =
                  [ni_min_abundance](const mcql::PeptideSp &peptide_sp) {
                    peptide_sp.get()->computeIsotopologues(ni_min_abundance);
                  };


              QFuture<void> res =
                QtConcurrent::map<std::vector<mcql::PeptideSp>::iterator>(
                  m_peptideStore.begin(),
                  m_peptideStore.end(),
                  mapComputeIsotopologues);
              res.waitForFinished();
            }

          for(std::pair<QString, mcql::MsRunPeptideListSp> msrun_pair :
              m_msfileList)
            {
              msrun_pair.second.get()->quantify(m_tmpDirName,
                                                m_retentionTimeRangeSeconds,
                                                m_uiMonitor,
                                                quantification_method,
                                                ni_min_abundance);


              writeQuantificationResults(
                quantify_id, withingroup, msrun_pair.second);
              msrun_pair.second.get()->flushChromatogramTraces();
            }
        }
    }
}

void
mcql::MassChroqMlParser::read_quantification_results()
{

  qDebug() << m_qxmlStreamReader.name();
  if(m_qxmlStreamReader.name().toString() == "quantification_results")
    {
      while(m_qxmlStreamReader.readNextStartElement())
        {
          if(m_qxmlStreamReader.name().toString() == "quantification_result")
            {
              mcql::QuantificationResultWriterInterfaceSp writer_sp =
                read_quantification_result();
              if(writer_sp.get() != nullptr)
                {
                  m_quantificationWriterList.push_back(writer_sp);
                }
            }
          else
            {

              m_qxmlStreamReader.raiseError(
                QObject::tr("quantification_results element has no %1 element")
                  .arg(m_qxmlStreamReader.name()));
              m_qxmlStreamReader.skipCurrentElement();
            }
        }
    }

  else
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("Not a quantification_results element"));
      m_qxmlStreamReader.skipCurrentElement();
    }
}

mcql::QuantificationResultWriterInterfaceSp
mcql::MassChroqMlParser::read_quantification_result()
{

  qDebug() << m_qxmlStreamReader.name();
  mcql::QuantificationResultWriterInterfaceSp writer_sp;
  if(m_qxmlStreamReader.name().toString() == "quantification_result")
    {
      QString output_file =
        m_qxmlStreamReader.attributes().value("output_file").toString();
      if(output_file.isEmpty())
        {
          m_qxmlStreamReader.raiseError(QObject::tr(
            "the quantification_result tag must have an output_file "
            "attribute."));
        }

      if(!ensureFilePathExists(output_file))
        {
          QDir file_dir(output_file);
          m_qxmlStreamReader.raiseError(
            QObject::tr("problem writing output to directory %1\n Please "
                        "check output file path and permissions")
              .arg(file_dir.absolutePath()));
        }

      QString format =
        m_qxmlStreamReader.attributes().value("format").toString();
      if(format.isEmpty())
        {
          m_qxmlStreamReader.raiseError(QObject::tr(
            "the quantification_result tag must have a format attribute."));
        }

      bool with_traces;
      QString xic_traces =
        m_qxmlStreamReader.attributes().value("xic_traces").toString();
      if(xic_traces == "true")
        {
          with_traces = true;
        }
      else
        {
          with_traces = false;
        }
      try
        {
          if(format == "gnumeric")
            {
              // return startGnumericResult(output_file);
            }
          else if(format == "xhtmltable")
            {
              // return startXhtmltableResult(output_file);
            }
          else if(format == "ods")
            {
              output_file = ensureFileExtension(output_file, "ods");
              writer_sp =
                std::make_shared<QuantificationResultOdsOutput>(output_file);
            }
          else if(format == "tsv")
            {
              output_file = ensureFileExtension(output_file, "d");
              writer_sp =
                std::make_shared<QuantificationResultTsvOutput>(output_file);
            }
          else if(format == "rdata")
            {

              m_qxmlStreamReader.raiseError(
                QObject::tr("output format 'rdata' is not available"));
            }
          else if(format == "masschroqml")
            {
              // startXmlResult(monitor_speed_list, output_file, with_traces);
            }
          else
            {
              m_qxmlStreamReader.raiseError(QObject::tr(
                "value '%1' of attribute format in quantification_result tag "
                "is "
                "not "
                "supported.\n Supported values are : ods, tsv and "
                "masschroqml."));
            }
        }
      catch(pappso::PappsoException &ods_error)
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("error creating result output file :\n%2")
              .arg(ods_error.qwhat()));
        }
      m_qxmlStreamReader.skipCurrentElement();
    }

  else
    {
      m_qxmlStreamReader.raiseError(
        QObject::tr("Not a quantification_result element"));
      m_qxmlStreamReader.skipCurrentElement();
    }
  qDebug() << m_qxmlStreamReader.name();
  return writer_sp;
}

bool
mcql::MassChroqMlParser::ensureFilePathExists(const QString &output_file)
{

  QFileInfo complete_path(output_file);
  //  qDebug()<< "MasschroqmlParser::ensureFilePathExists complete_path.dir()"
  //  << complete_path.dir().absolutePath() ;
  // qDebug() << QFileInfo("~/examples/191697/main.cpp").dir().absolutePath();

  // qDebug() << QFileInfo("~/examples/191697/main").dir().absolutePath();
  QDir only_dir(complete_path.dir().absolutePath());
  if(only_dir.exists())
    {
      return true;
    }
  else
    {
      // try to create it
      if(only_dir.mkpath(only_dir.absolutePath()) == false)
        {
          return false;
          qDebug() << "MasschroqmlParser::ensureFilePathExists FALSE "
                   << only_dir.absolutePath();
        }
      else
        {
          return true;
        }
    }
  return false;
}

QString
mcql::MassChroqMlParser::ensureFileExtension(const QString &output_file,
                                             const QString &file_extension)
{
  QFileInfo filename(output_file);
  if(filename.suffix().isEmpty())
    {
      return QString(output_file + "." + file_extension);
    }
  return output_file;
}

void
mcql::MassChroqMlParser::writeQuantificationResults(
  const QString &quantify_id,
  const QString &withingroup,
  const mcql::MsRunPeptideListSp &msrun_peptide_list) const
{
  for(auto &writer_sp : m_quantificationWriterList)
    {

      msrun_peptide_list.get()->writeQuantificationResults(
        m_uiMonitor, writer_sp, quantify_id, withingroup);
    }
}

void
mcql::MassChroqMlParser::read_protein()
{

  //<protein desc="conta|P02769|ALBU_BOVIN SERUM ALBUMIN PRECURSOR."
  // id="P1.1" />
  qDebug() << m_qxmlStreamReader.name();
  if(m_qxmlStreamReader.name().toString() == "protein")
    {
      QString id   = m_qxmlStreamReader.attributes().value("id").toString();
      QString desc = m_qxmlStreamReader.attributes().value("desc").toString();
      if(id.isEmpty())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("the protein tag must have an id attribute."));
        }
      /// create a new Protein object and set its description
      mcql::ProteinSp p_protein;
      try
        {
          p_protein = std::make_shared<mcql::Protein>(id, desc);
        }
      catch(pappso::PappsoException &error)
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("problem creating protein :\n%1").arg(error.what()));
        }
      /// add this protein to _p_proteins (: map<id, Protein *>)
      m_proteinMap.insert({id, p_protein});
      qDebug();
      m_qxmlStreamReader.skipCurrentElement();
    }

  else
    {
      m_qxmlStreamReader.raiseError(QObject::tr("Not a protein element"));
      m_qxmlStreamReader.skipCurrentElement();
    }
  qDebug() << m_qxmlStreamReader.name();
}

void
mcql::MassChroqMlParser::read_group()
{
  //<group id="all_msrun" data_ids="msruna1 msruna2 msruna3 msruna4
  // msruna5 msruna6 msruna7 msruna8 msruna9 msrunb10 msrunb11 msrunb12
  // msrunb13 msrunb14 msrunb15 msrunb16 msrunb17 msrunb18 msrunb19"/>

  qDebug() << m_qxmlStreamReader.name();
  if(m_qxmlStreamReader.name().toString() == "group")
    {
      QString id = m_qxmlStreamReader.attributes().value("id").toString();
      QStringList msrun_id_str_list =
        m_qxmlStreamReader.attributes()
          .value("data_ids")
          .toString()
          .split(" ", Qt::SplitBehaviorFlags::SkipEmptyParts);
      if(id.isEmpty())
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("the group tag must have an id attribute."));
        }
      /// create a new Protein object and set its description
      mcql::MsRunGroupSp group_sp;
      std::vector<mcql::MsRunPeptideListSp> msrun_list;
      if(msrun_id_str_list.size() == 0)
        {
          m_qxmlStreamReader.raiseError(
            QObject::tr("the group tag must have a data_ids attribute."));
        }
      else
        {
          for(const QString &msrun_id : msrun_id_str_list)
            {
              auto it = m_msfileList.find(msrun_id);
              if(it != m_msfileList.end())
                {
                  msrun_list.push_back(it->second);
                }
              else
                {
                  m_qxmlStreamReader.raiseError(
                    QObject::tr("msrun id %1 not found").arg(msrun_id));
                }
            }
        }
      group_sp = std::make_shared<mcql::MsRunGroup>(id, msrun_list);
      /// add this protein to _p_proteins (: map<id, Protein *>)
      m_msRunGroupSpList.insert({id, group_sp});
      qDebug();
      m_qxmlStreamReader.skipCurrentElement();
    }

  else
    {
      m_qxmlStreamReader.raiseError(QObject::tr("Not a group element"));
      m_qxmlStreamReader.skipCurrentElement();
    }
}
