/**
 * \file mcql/core/peptide.cpp
 * \date 24/09/2024
 * \author Olivier Langella
 * \brief peptide model in masschroqlite
 */

/*******************************************************************************
 * Copyright (c) 2024 Olivier Langella
 *<Olivier.Langella@universite-paris-saclay.fr>.
 *
 * This file is part of MassChroQ.
 *
 *     MassChroQ is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MassChroQ is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MassChroQ.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

#include "peptide.h"

namespace mcql
{
Peptide::Peptide(const QString &id,
                 pappso::PeptideSp peptide_sp,
                 const std::vector<ProteinSp> &protein_list)
  : m_id(id), m_proteinSpList(protein_list)
{
  msp_peptide = peptide_sp;
}

Peptide::Peptide(const Peptide &other)
  : m_id(other.m_id), m_proteinSpList(other.m_proteinSpList)
{
  msp_peptide = other.msp_peptide;
  m_mods      = other.m_mods;
}

Peptide::~Peptide()
{
  if(mpa_peptideNaturalIsotopeList != nullptr)
    delete mpa_peptideNaturalIsotopeList;
}

} // namespace mcql

const pappso::PeptideSp &
mcql::Peptide::getPappsoPeptideSp() const
{
  return msp_peptide;
}

void
mcql::Peptide::computeIsotopologues(double ni_min_abundance)
{

  if(ni_min_abundance > 0)
    {

      qDebug() << "if (_minimum_isotope_abundance > 0)";
      mpa_peptideNaturalIsotopeList =
        new pappso::PeptideNaturalIsotopeList(msp_peptide, 0.001);
    }
}

const pappso::PeptideNaturalIsotopeList *
mcql::Peptide::getPeptideNaturalIsotopeList() const
{
  return mpa_peptideNaturalIsotopeList;
}

const QString &
mcql::Peptide::getId() const
{
  return m_id;
}

void
mcql::Peptide::setMods(const QString &mods)
{
  m_mods = mods;
}

const QString &
mcql::Peptide::getMods() const
{
  return m_mods;
}

const std::vector<pappso::PeptideNaturalIsotopeAverageSp> &
mcql::Peptide::getPeptideNaturalIsotopeAverageSpList(
  pappso::PrecisionPtr precision, std::uint8_t charge, double ni_min_abundance)
{

  QMutexLocker lock(&m_mutex);
  if(m_niMinAbundance == ni_min_abundance)
    {
    }
  else
    {
      m_peptideNaturalIsotopeAverageSpListByCharge.clear();
      m_niMinAbundance = ni_min_abundance;
    }

  auto it = m_peptideNaturalIsotopeAverageSpListByCharge.find(charge);
  if(it != m_peptideNaturalIsotopeAverageSpListByCharge.end())
    {
      if(it->second.size() == 0)
        {
          it->second = mpa_peptideNaturalIsotopeList->getByIntensityRatio(
            charge, precision, ni_min_abundance);
        }

      return it->second;
    }
  else
    {
      auto it_insert = m_peptideNaturalIsotopeAverageSpListByCharge.insert(
        {charge,
         mpa_peptideNaturalIsotopeList->getByIntensityRatio(
           charge, precision, ni_min_abundance)});

      return it_insert.first->second;
    }
}

void
mcql::Peptide::addObservedChargeState(std::uint8_t charge)
{
  if(m_allObservedChargeStateList.end() ==
     std::find(m_allObservedChargeStateList.begin(),
               m_allObservedChargeStateList.end(),
               charge))
    {
      m_allObservedChargeStateList.push_back(charge);
    }
}

const std::vector<std::uint8_t> &
mcql::Peptide::getAllObservedChargeStateList() const
{
  return m_allObservedChargeStateList;
}

void
mcql::Peptide::addNotObservedInMsRunSp(const mcql::MsRunSp &msrun_sp)
{
  m_notObservedInMsRunSpList.push_back(msrun_sp);
}
