/**
 * \file mcql/core/peptide.h
 * \date 24/09/2024
 * \author Olivier Langella
 * \brief peptide model in masschroqlite
 */

/*******************************************************************************
 * Copyright (c) 2024 Olivier Langella
 *<Olivier.Langella@universite-paris-saclay.fr>.
 *
 * This file is part of MassChroQ.
 *
 *     MassChroQ is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     MassChroQ is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with MassChroQ.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

#pragma once


#include <pappsomspp/peptide/peptidenaturalisotopelist.h>
#include "msrun.h"
#include "protein.h"
#include <cstdint>

namespace mcql
{

class Peptide;

typedef std::shared_ptr<Peptide> PeptideSp;

/**
 * unique studied entity representing the molecule we want to measure
 */
class Peptide
{
  public:
  /**
   * Default constructor
   */
  Peptide(const QString &id,
          pappso::PeptideSp msp_peptide,
          const std::vector<ProteinSp> &protein_list);

  /**
   * Copy constructor
   *
   * @param other TODO
   */
  Peptide(const Peptide &other);

  /**
   * Destructor
   */
  virtual ~Peptide();

  const pappso::PeptideSp &getPappsoPeptideSp() const;

  void computeIsotopologues(double ni_min_abundance);

  const pappso::PeptideNaturalIsotopeList *getPeptideNaturalIsotopeList() const;

  const QString &getId() const;

  void setMods(const QString &mods);
  const QString &getMods() const;

  const std::vector<pappso::PeptideNaturalIsotopeAverageSp> &
  getPeptideNaturalIsotopeAverageSpList(pappso::PrecisionPtr precision,
                                        std::uint8_t charge,
                                        double ni_min_abundance);

  void addObservedChargeState(std::uint8_t charge);
  void addNotObservedInMsRunSp(const mcql::MsRunSp &msrun_sp);

  const std::vector<std::uint8_t> &getAllObservedChargeStateList() const;

  private:
  const QString m_id;
  const std::vector<ProteinSp> m_proteinSpList;
  QString m_mods;
  pappso::PeptideSp msp_peptide;
  pappso::PeptideNaturalIsotopeList *mpa_peptideNaturalIsotopeList = nullptr;
  std::map<std::uint8_t, std::vector<pappso::PeptideNaturalIsotopeAverageSp>>
    m_peptideNaturalIsotopeAverageSpListByCharge;
  double m_niMinAbundance = 0;
  std::vector<std::uint8_t> m_allObservedChargeStateList;
  std::vector<mcql::MsRunSp> m_notObservedInMsRunSpList;

  QMutex m_mutex;
};
} // namespace mcql
